/*
 * Copyright (C) 1997.  Netscape Communications Corp. 
 *
 * Unsupported Example Code.
 * 
 * All examination and use of this code is covered by the
 * unsupported software license agreement accompanying it.
 *
 */

/*
 * pblock.h: Header for Parameter Block handling functions
 *
 * This is a version of the pblock interface specifically for use in
 * VGI programs.
 *
 * A parameter block is a set of name=value pairs which are generally used 
 * as parameters, but can be anything. They are kept in a hash table for 
 * reasonable speed, but if you are doing any intensive modification or
 * access of them you should probably make a local copy of each parameter
 * while working.
 *
 * When creating a pblock, you specify the hash table size for that pblock.
 * You should set this size larger if you know that many items will be in 
 * that pblock, and smaller if only a few will be used or if speed is not
 * a concern.
 *
 * The hashing function is very simple right now, and only looks at the 
 * first character of name.
 *
 */

#ifndef ns_pblock_h
#define ns_pblock_h

#ifdef __cplusplus
extern "C" {
#endif

/* ------------------------------ Structures ------------------------------ */

typedef struct {
    char *name;
    char *value;
} pb_param;

struct pb_entry {
    pb_param *param;
    struct pb_entry *next;
};

typedef struct {
    int hsize;
    struct pb_entry **ht;
} pblock;


/* ------------------------------ Prototypes ------------------------------ */


/*
 * param_create creates a parameter with the given name and value. If name
 * and/or value are non-NULL, they are *copied* and placed into the new
 * pb_param struct.
 */

pb_param *param_create(char *name, char *value);

/*
 * param_free frees a given parameter if it's non-NULL, and returns 1 if
 * p was non-NULL, and 0 if p was NULL.
 * 
 * Useful for error checking pblock_remove.
 */

int param_free(pb_param *pp);

/* 
 * pblock_create creates a new pblock with hash table size n.
 * 
 * It returns the newly allocated pblock, or NULL on failure.
 */

pblock *pblock_create(int n);

/*
 * pblock_free frees the given pblock and any entries inside it.
 * 
 * If you want to save anything in a pblock, remove its entities with 
 * pblock_remove first and save the pointers you get.
 */

void pblock_free(pblock *pb);

/* 
 * pblock_find finds the entry with the given name in pblock pb.
 *
 * If it is successful, it returns the param block. If not, it returns NULL.
 */

pb_param *pblock_find(char *name, pblock *pb);

/*
 * pblock_remove behaves exactly like pblock_find, but removes the given
 * entry from pb.
 */

pb_param *pblock_remove(char *name, pblock *pb);


/*
 * pblock_findval finds the entry with the given name in pblock pb, and
 * returns its value, otherwise returns NULL.
 */

char *pblock_findval(char *name, pblock *pb);

/*
 * pblock_nvinsert creates a new parameter with the given name and value.
 * and inserts it into pblock pb. The name and value are copied.  They must
 * both be non-NULL.   Returns the pb_param it allocated (in case you 
 * need it).  The returned pointer is NULL if allocation failed.
 *
 * If a parameter with the given value exists in the pblock, it is replaced.
 */

pb_param *pblock_nvinsert(char *name, char *value, pblock *pb);


/*
 * pblock_pinsert inserts a pb_param into a pblock.  The contained name
 * and value must be non-NULL.  Returns 1 on success, 0 on failure.
 */

int pblock_pinsert(pb_param *pp, pblock *pb);

/*
 * pblock_copy copies the entries in the given source pblock to the 
 * destination one. The entries are newly allocated so that the original
 * pblock may be freed or the new one changed without affecting the other.
 * Returns 1 on success, 0 on failure.
 */

/* NOTE the unusual order of src, dst */
int pblock_copy(pblock *src, pblock *dst);

/*
 * pblock_dup creates a new pblock and copies the given source pblock
 * into it.  The entries are newly allocated so that the original pblock
 * may be freed or the new one changed without affecting the other.
 * Returns a pointer to the new pblock on success, or NULL on failure.
 */

pblock *pblock_dup(pblock *src);


/*
 * ------------------------------ Iterator -------------------------------
 *
 * This provides a way of iterating over the entries in the pblock.
 * The iterator traverses the entries in no guaranteed order.
 * Adding or removing entries in the pblock while using an iterator
 * renders the iterator invalid.  It's usage becomes undefined and
 * dangerous if you do that.
 * 
 * Here is an example of usage iterating over the entries in the
 * pblock indicated by 'the_pblock';
 *
 * pblock *the_pblock;  -- points to a pblock we want to iterate over
 * ... 
 * pb_iter i;
 * for (pb_iter_init(&i, the_pblock); !pb_iter_at_end(&i); pb_iter_next(&i)) {
 *    pb_param *pp = pb_iter_cur(&i);
 *    ... here you can do something that uses pp->name, pp->value
 * }  
 *
 */

typedef struct {
    pblock *pb;
    int b;
    struct pb_entry *p;
} pb_iter;


/* Create iterator */
void pb_iter_init(pb_iter *i, pblock *pb);

/* Returns 1 if at end of iteration, 0 otherwise */
int pb_iter_at_end(pb_iter *i);

/* Returns parameter at current position */
pb_param *pb_iter_cur(pb_iter *i);

/* Advances iteration position */
void pb_iter_next(pb_iter *i);


#ifdef __cplusplus
}
#endif

#endif


