/*
 * Copyright (C) 1997.  Netscape Communications Corp. 
 *
 * Unsupported Example Code.
 * 
 * All examination and use of this code is covered by the
 * unsupported software license agreement accompanying it.
 *
 */

/*
 * This is example code for a simple VGI program that
 * will authenticate users based on an existing Unix password source
 * scheme on the certificate server host.  It is an example for
 * illustrative purposes only.   This should not be considered deployment
 * quality code nor an indication of any recommendation for a deployment.
 * 
 * Note in particular that it is not recommended to use YP (NIS)
 * as a password source on a secure certificate server host, and
 * that in an actual deployment it is advisable to use a different
 * source of password information than that used for login on the
 * certificate server host machine.
 *
 */

/*
 * On most Unix systems this will compile using
 *         % cc -o unixexample unixexample.c -I../include -L../lib -lvgiutil
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <ctype.h>
#include <pwd.h>
#include <crypt.h>

#include "pblock.h"
#include "urlcod.h"


char *progName = 0;


/*
 * This routine is used to send a crude error message back to the client.
 *
 * In actual practice you can send arbitrary HTML and better error messages.
 *
 */

void
gatewayError(const char *msg)
{
    printf("Content-Type: text/html\n");
    printf("Content-Length: %d\n", strlen(progName) + 2 + strlen(msg));
    printf("\n");
    printf("%s: %s", progName, msg);
    exit(0);
}    


/* Arbitrary -- you can use allocation on heap (malloc()) to avoid this */ 
#define INBUFSIZE 8192

/* Varies -- may be in limits.h.  Maximum allowed length of a user name */
#ifndef LOGNAME_MAX
#define LOGNAME_MAX 8
#endif

/* Maximum length of a subject name */
#define SUBJECT_MAX 255
#define REALNAME_MAX 64

/* Validity period for certs issued - 6 months (180 days) */
#define DAY 86400
#define VALIDITY_PERIOD (6*30*DAY)


int
main(int argc, char **argv, char **envp)
{
    char *contentLenStr = getenv("CONTENT_LENGTH");
    char inBuf[INBUFSIZE];
    int contentLen;
    int actualLen;
    pblock *inArgs;
    pblock *outArgs;
    char *usernameRaw;
    char username[LOGNAME_MAX+1];
    char realname[REALNAME_MAX+1];
    char *password;
    struct passwd *pwent;
    char subject[SUBJECT_MAX];
    char validityBuf[64];
    char *spki;
    char *outForm;

    progName = argv[0];

    if (contentLenStr == 0) {
	printf("This program is a VGI program; it expects to be\n"
	       "run through the certificate server auto-verification\n"
	       "interface\n");
	exit(1);
    }
    
    contentLen = atoi(contentLenStr);
    if (contentLen <= 0) {
	gatewayError("Ill-formed contentlength from gateway.");
    }
    if (contentLen > INBUFSIZE-1) {
	gatewayError("Size of submission is too large");
    }
    

    actualLen = 0;
    while (actualLen < contentLen) {
	int c = getc(stdin);
	if (c == EOF) break;
	inBuf[actualLen] = (char) c;
	actualLen++;
    }
    inBuf[actualLen] = 0;    

    if (actualLen < contentLen) {
	gatewayError("Actual length less than stated content length.");
    }
    
    inArgs = url_encoding_to_pblock(inBuf);
    if (inArgs == 0) {
	gatewayError("Input to VGI program was improperly formed.");
    }

    /*
     * Map input to a subject name and authenticate as you wish.
     * Here the form submission must include a "username" and "password"
     * and we use the machine's source of user password information
     * to authenticate and formulate the subject name.
     */

    usernameRaw = pblock_findval("username", inArgs);
    if (usernameRaw == 0) {
	gatewayError("Rejected. No username in request.");
    }
    if (strlen(usernameRaw) < 1 || strlen(usernameRaw) > LOGNAME_MAX) {
	gatewayError("Rejected.   Username is empty or too long.");
    }
    {
	/* convert user name to lower case */
	char *s, *t;
	for (s = usernameRaw, t=&username[0]; *s != 0; s++, t++ ) {
	    *t = (char) tolower(*s);
	}
	*t = 0;
    }
    
    password = pblock_findval("passwd", inArgs);
    if (username == 0) {
	gatewayError("Rejected. No username in request.");
    }

    spki = pblock_findval("subjectPublicKeyInfo",inArgs);
    if (spki == 0) {
	gatewayError("Rejected.  No public key info in request.");
    }


    /*
     * Sites that use shadow password files should use getspnam here
     * and will have to deal with issues of permissions to access
     * the shadow file by the server.
     */
    
    pwent = getpwnam(username);
    if (pwent == 0) {
	/* intentionally non-specific -- you may wish to say more */
	gatewayError("The username or password is incorrect.  "
		     "Check that you have entered these correctly");
    }


    if (strcmp(crypt(password, pwent->pw_passwd), pwent->pw_passwd)) {
	/* bad password */
	/* intentionally non-specific -- you may wish to say more */
	gatewayError("The username or password is incorrect.  "
		     "Check that you have entered these correctly.");
    }


    /* Get the real name from the gecos field */
    {
	char *s, *t;
	int len;
	for (s=pwent->pw_gecos, t=&realname[0], len=0; *s != 0 && *s != ',' &&
	     len < REALNAME_MAX; s++, t++, len++) {
	    *t = *s;
	}
	*t = 0;
    }
	
    /* Allocate pblock for holding parameters of issue operation */
    outArgs = pblock_create(16);
    if (outArgs == 0) {
	gatewayError("Unable to create pblock.");
    }

    /*
     * Form subject name from user's login, real name, and a
     * fixed organization name.  This should be done in a way that
     * matches your directory structure.  The name formulation here 
     * may or may not match your organization.  You should change
     * it so that it does.  You can do fancier things like determine
     * an organization unit (OU) from the user's primary group id or
     * you can use the LDAP SDK itself and use an existing password
     * in the directory for the initial authentication.
     */
    {
	/*
	 * Note -- this will fit in subject[] without danger of overflow.
	 * If you change this you may have to check it.
	 */
	char *organization = "Test Corporation";
	char *country = "US";
	char *subjectFormat = "UID=%s, CN=%s, O=%s, C=%-2s";
	sprintf(subject, subjectFormat,
		username,  realname, organization, country);
    }

    /* See the documentation for more information on how to formulate
     * an issue request and how to specify additional extensions.
     */

    /* validity */
    sprintf(validityBuf, "%d", VALIDITY_PERIOD);

   

    /* Generate parameters for issue request */
    if (!pblock_nvinsert("op", "issue", outArgs) ||
	!pblock_nvinsert("subjectPublicKeyInfo", spki, outArgs) ||
	!pblock_nvinsert("subject", subject, outArgs) ||

	/* validity */
	!pblock_nvinsert("validityLength", validityBuf, outArgs) ||

	/* instruct server to add these extensions if extensions are enabled */
	!pblock_nvinsert("AuthorityKeyIdentifier", "", outArgs) ||
	!pblock_nvinsert("certTypeSSLClient", "true", outArgs) ||
	!pblock_nvinsert("certTypeEmail", "true", outArgs)) {

	/* failed unexpectedly */
	gatewayError("Failure while setting up issue parameters.");
    }
	
    /* URL-encode the parameters */
    outForm = pblock_to_url_encoding(outArgs);
    if (outForm == 0) {
	gatewayError("Failed url-encoding issue request.");
    }

    /* Output it back to the server */
    printf("Content-Type: application/x-www-form-urlencoded\n");
    printf("Content-Length: %d\n", strlen(outForm));
    printf("\n");
    printf("%s", outForm);

    exit(0);
}
