/*
 * Copyright (C) 1997.  Netscape Communications Corp. 
 *
 * Unsupported Example Code.
 * 
 * All examination and use of this code is covered by the
 * unsupported software license agreement accompanying it.
 *
 */

/*
 * pblock.c: Handles Parameter Blocks
 * 
 * See pblock.h for public documentation.
 *
 */


#include "pblock.h"
#include <string.h>
#include <stdlib.h>

/* FREE must be ANSI-compliant (FREE(0) is a no-op) */
#define MALLOC malloc
#define FREE free
#define STRDUP strdup


/* ----------------------------- param_create ----------------------------- */


pb_param *param_create(char *name, char *value)
{
    pb_param *ret = (pb_param *) MALLOC(sizeof(pb_param));
    if (ret==0) {
	return 0;
    }

    ret->name = 0;
    ret->value = 0;
    if (name) {
	ret->name = STRDUP(name);
	if (ret->name == 0) {
	    param_free(ret);
	    return 0;
	}
    }
    if (value) {
	ret->value = STRDUP(value);
	if (ret->value == 0) {
	    param_free(ret);
	    return 0;
	}
    }
    return ret;
}


/* ------------------------------ param_free ------------------------------ */


int param_free(pb_param *pp) {
    if(pp) {
        FREE(pp->name);
        FREE(pp->value);
        FREE(pp);
        return 1;
    }
    return 0;
}


/* ---------------------------- pblock_create ----------------------------- */

pblock *pblock_create(int n) {
    pblock *pb = (pblock *) MALLOC(sizeof(pblock));
    register int x;

    if (pb == 0) {
	return 0;
    }

    pb->hsize = n;
    pb->ht = (struct pb_entry **) MALLOC(n * sizeof(struct pb_entry *));
    if (pb->ht == 0) {
	pblock_free(pb);
	return 0;
    }
    for(x=0;x<n;x++) pb->ht[x] = NULL;

    return pb;
}


/* ----------------------------- pblock_free ------------------------------ */


void pblock_free(pblock *pb) {

    if (pb->ht) {
	register int x,n;
	for(x = 0,n = pb->hsize; x<n; x++) {
	    struct pb_entry *p,*q;
	    p = pb->ht[x];
	    while(p) {
		q = p->next;
		param_free(p->param);
		FREE(p);
		p = q;
	    }
	}
	FREE(pb->ht);
    }
    FREE(pb);
}


/* ---------------------------- pblock_pinsert ---------------------------- */


#define PBLOCK_HASH(name,pb) ( ((unsigned int)(name[0])) % (pb)->hsize)

int pblock_pinsert(pb_param *pp, pblock *pb)
{
    int ph;
    struct pb_entry *newOne;
    if (pp->name == 0 || pp->value == 0) {
	/* name and value must be non-null */
	return 0;
    }
    ph = PBLOCK_HASH(pp->name,pb);
    newOne = (struct pb_entry *) MALLOC(sizeof(struct pb_entry));
    if (newOne==0) {
	return 0;
    }
	
    newOne->param = pp;
    newOne->next = pb->ht[ph];
    pb->ht[ph] = newOne;
    return 1;
}


/* --------------------------- pblock_nvinsert ---------------------------- */


pb_param *pblock_nvinsert(char *name, char *value, pblock *pb) {

    pb_param *pp = 0;
    if (value == 0) {
	/* both name and value must be non-null */
	return 0;
    }
    pp = pblock_find(name, pb);
    if (pp) {
	/* named parameter exists */
	char *newValue = STRDUP(value);
	if (newValue == 0) {
	    return 0;
	}
	FREE(pp->value);
	pp->value = newValue;
	return pp;
    }
    /* named parameter does not exist */
    pp = param_create(name, value);
    if (pp == 0) {
	return 0;
    }
    if (!pblock_pinsert(pp,pb)) {
	param_free(pp);
	return 0;
    }
    return pp;
}



/* ---------------------------- pblock_findval ---------------------------- */


char *pblock_findval(char *name, pblock *pb) {
    pb_param *pp = pblock_find(name, pb);

    return (pp ? pp->value : NULL);
}





/* ------------------------------ _pblock_fr ------------------------------ */


static
pb_param *_pblock_fr(char *name, pblock *pb, int remove) {
    struct pb_entry *p,*q;
    int ph = PBLOCK_HASH(name,pb);

    for(p = pb->ht[ph], q = NULL; p; q = p, p = p->next) {
        if(!strcmp(name,p->param->name)) {
            pb_param *pp = p->param;

            if(remove) {
                if(q) 
                    q->next = p->next;
                else  
                    pb->ht[ph] = p->next;

                FREE(p);
            }
            return pp;
        }
    }
    return NULL;
}


pb_param *pblock_find(char *name, pblock *pb) {
    return _pblock_fr(name, pb, 0);
}
	
pb_param *pblock_remove(char *name, pblock *pb) {
    return _pblock_fr(name, pb, 1);
}



/* ----------------------------- pblock_copy ------------------------------ */


int pblock_copy(pblock *src, pblock *dst)
{
    register int x,n;
    struct pb_entry *p;

    for(x = 0,n = src->hsize; x < n; x++) {
        for(p = src->ht[x]; p; p = p->next) {
	    if (pblock_nvinsert(p->param->name, p->param->value, dst) == 0) {
		return 0;
	    }
	}
    }
}

/* ---------------------------- pblock_dup -------------------------------- */

pblock *pblock_dup(pblock *src)
{
    pblock *dst = 0;

    if (!src)
        return NULL;

    dst = pblock_create(src->hsize);
    if (dst == 0) {
	return 0;
    }

    if (!pblock_copy(src, dst)) {
	pblock_free(dst);
	return 0;
    }

    return dst;
}


/* ------------------------------ Iterator ------------------------------- */


/* Create iterator */
void pb_iter_init(pb_iter *i, pblock *pb)
{
    i->pb = pb;
    i->b = 0;
    i->p = pb->ht[0];
    while (i->p == 0 && i->b < i->pb->hsize - 1) {
	i->b++;
	i->p = i->pb->ht[i->b];
    }
}


/* Returns 1 if at end of iteration, 0 otherwise */
int pb_iter_at_end(pb_iter *i) {
    return (i->p == 0);
}

/* Returns parameter at current position */
pb_param *pb_iter_cur(pb_iter *i) {
    return (i->p->param);
}

/* Advances iteration position */
void pb_iter_next(pb_iter *i) {
    if (!pb_iter_at_end(i)) {
	i->p = i->p->next;
	while (i->p == 0 && i->b < i->pb->hsize - 1) {
	    i->b++;
	    i->p = i->pb->ht[i->b];
	}
    }
}
